"""
Copyright 2019 Marco Lattuada
Copyright 2021 Bruno Guindani

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import abc
import copy
import itertools
import pprint
import random
import sys

import custom_logger

import data_preparation.normalization
import data_preparation.random_splitting
import data_preparation.xgboost_feature_selection

import model_building.decision_tree_experiment_configuration as dt
import model_building.experiment_configuration as ec
import model_building.lr_ridge_experiment_configuration as lr
import model_building.nnls_experiment_configuration as nnls
import model_building.random_forest_experiment_configuration as rf
import model_building.stepwise_experiment_configuration as sw
import model_building.svr_experiment_configuration as svr
import model_building.xgboost_experiment_configuration as xgb
import model_building.wrapper_experiment_configuration as wec


class ExpConfsGenerator(abc.ABC):
    """
    Abstract class representing a generators of experiment configurations

    Its descendants are hierarchically used to generate all the experiment configurations to be evaluated.
    Evaluation is not triggered by generators but directly by ModelBuilding

    Attributes
    ----------
    _campaign_configuration: dict of str: dict of str: str
        The set of options specified by the user though command line and campaign configuration files

    _random_generator: RandomGenerator
        The random generator used to generate random numbers

    _experiment_configurations: list of ExperimentConfiguration
        The list of ExperimentConfiguration created

    _logger: Logger
        The logger associated with this class and its descendents

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator
        """

        self._experiment_configurations = []
        self._campaign_configuration = campaign_configuration
        self._random_generator = random.Random(seed)
        self._logger = custom_logger.getLogger(__name__)

    @abc.abstractmethod
    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generates the set of experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used for the signature of the ExperimentConfiguration generated by this generator

        regression_inputs: RegressionInputs
            The data to be used by the ExperimentConfiguration

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """

    def __deepcopy__(self, memo):
        raise NotImplementedError()


class MultiExpConfsGenerator(ExpConfsGenerator):
    """
    ExpConfsGenerator which wraps together multiple generators

    Attributes
    ----------
    generators: dict of str: ExpConfsGenerator
        The ExpConfsGenerator objects to be used

    Methods
    -------
    _get_deep_copy_parameters()
        Computes the parameters to be used by __deepcopy__ of subclasses
    """

    def __init__(self, campaign_configuration, seed, generators):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used in random based activities

        generators: dict of str: ExpConfsGenerator
            The ExpConfsGenerator objects to be used
        """
        assert generators
        super().__init__(campaign_configuration, seed)
        self._generators = generators

    def __deepcopy__(self, memo):
        raise NotImplementedError()

    def _get_deep_copy_parameters(self):
        """
        Returns deep copy of relevant parameters, but with a different random seed

        Returns
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            A randomized seed to be used in random based activities

        generators: dict of str: ExpConfsGenerator
            The ExpConfsGenerator objects used
        """
        campaign_configuration = self._campaign_configuration
        seed = self._random_generator.random()
        generators = copy.deepcopy(self._generators)
        return campaign_configuration, seed, generators


class MultiTechniquesExpConfsGenerator(MultiExpConfsGenerator):
    """
    ExpConfsGenerator which wraps together the ExpConfsGenerator of the single techniques

    This class wraps the single TechniqueExpConfsGenerator instances which refer to the single techinique

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generates the set of experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used for the signature of the ExperimentConfiguration generated by this generator

        regression_inputs: RegressionInputs
            The data to be used by the ExperimentConfiguration

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by MultiTechniquesExpConfsGenerator")
        return_list = []
        assert self._generators

        for key, generator in self._generators.items():
            new_prefix = prefix.copy()
            new_prefix.append(key)
            assert new_prefix
            return_list.extend(generator.generate_experiment_configurations(new_prefix, regression_inputs))

        assert return_list
        self._logger.info("<--")
        return return_list

    def __deepcopy__(self, memo):
        campaign_configuration, seed, generators = self._get_deep_copy_parameters()
        return MultiTechniquesExpConfsGenerator(campaign_configuration, seed, generators)


class TechniqueExpConfsGenerator(ExpConfsGenerator):
    """
    Class which generalize classes for generate points to be explored for each technique

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, technique):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            the seed to be used in the generation (ignored in this class)

        technique: ec.Technique
            the technique to be used in the generation
        """

        super().__init__(campaign_configuration, seed)
        self._technique = technique

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by TechniqueExpConfsGenerator %s", ec.enum_to_configuration_label[self._technique])
        assert prefix
        first_key = ""
        # We expect that hyperparameters for a technique are stored in campaign_configuration[first_key]
        # as a dictionary from string to list of values
        if self._technique == ec.Technique.NONE:
            self._logger.error("Not supported regression technique")
            sys.exit(-1)

        first_key = ec.enum_to_configuration_label[self._technique]
        hyperparams = self._campaign_configuration[first_key]
        self._logger.debug("Hyperparams are %s", pprint.pformat(hyperparams, width=1))
        hyperparams_names = []
        hyperparams_values = []
        # Adding dummy hyperparameter to have at least two hyperparameters
        hyperparams_names.append('dummy')
        hyperparams_values.append([0])

        self._logger.debug("Computing set of hyperparameters combinations")
        for hyperparam in hyperparams:
            hyperparams_names.append(hyperparam)
            hyperparams_values.append(hyperparams[hyperparam])

        assert not self._experiment_configurations

        # Cartesian product of parameters
        for combination in itertools.product(*hyperparams_values):
            hyperparams_point_values = {}
            for hyperparams_name, hyperparams_value in zip(hyperparams_names, combination):
                hyperparams_point_values[hyperparams_name] = hyperparams_value

            if self._technique == ec.Technique.LR_RIDGE:
                point = lr.LRRidgeExperimentConfiguration(self._campaign_configuration, hyperparams_point_values,
                                                          regression_inputs, prefix)
            elif self._technique == ec.Technique.XGBOOST:
                point = xgb.XGBoostExperimentConfiguration(self._campaign_configuration, hyperparams_point_values,
                                                           regression_inputs, prefix)
            elif self._technique == ec.Technique.DT:
                point = dt.DecisionTreeExperimentConfiguration(self._campaign_configuration, hyperparams_point_values,
                                                               regression_inputs, prefix)
            elif self._technique == ec.Technique.RF:
                point = rf.RandomForestExperimentConfiguration(self._campaign_configuration, hyperparams_point_values,
                                                               regression_inputs, prefix)
            elif self._technique == ec.Technique.SVR:
                point = svr.SVRExperimentConfiguration(self._campaign_configuration, hyperparams_point_values,
                                                       regression_inputs, prefix)
            elif self._technique == ec.Technique.NNLS:
                point = nnls.NNLSExperimentConfiguration(self._campaign_configuration, hyperparams_point_values,
                                                         regression_inputs, prefix)
            elif self._technique == ec.Technique.STEPWISE:
                point = sw.StepwiseExperimentConfiguration(self._campaign_configuration, hyperparams_point_values,
                                                           regression_inputs, prefix)
            else:
                self._logger.error("Not supported regression technique")
                sys.exit(-1)

            self._experiment_configurations.append(point)

        assert self._experiment_configurations

        self._logger.info("<--")

        return self._experiment_configurations

    def __deepcopy__(self, memo):
        ret = TechniqueExpConfsGenerator(self._campaign_configuration, self._random_generator.random(), self._technique)
        return ret


class RepeatedExpConfsGenerator(MultiExpConfsGenerator):
    """
    Invokes n times the wrapped ExpConfsGenerator with n different seeds

    Creates n different copies of the wrapped generator which differs for the seed used to generate random numbers

    Attributes
    ----------
    _repetitions_number: integer
        The number of different seeds to be used

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, repetitions_number, wrapped_generator):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed used to initialize the random number generator

        repetitions_number: integer
            The number of different copies to be created

        wrapped_generator: ExpConfsGenerator
            The wrapped generator to be invoked
        """

        self._repetitions_number = repetitions_number

        wrapped_generators = {"run_0": wrapped_generator}

        for run_index in range(1, self._repetitions_number):
            wrapped_generators["run_" + str(run_index)] = copy.deepcopy(wrapped_generator)

        super().__init__(campaign_configuration, seed, wrapped_generators)

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by RepeatedExpConfsGenerator")
        return_list = []
        assert self._generators

        for key, generator in self._generators.items():
            new_prefix = prefix.copy()
            new_prefix.append(key)
            self._logger.info("-->Generating experiments for run %s", str(key))
            return_list.extend(generator.generate_experiment_configurations(new_prefix, regression_inputs))
            self._logger.info("<--")

        assert return_list
        self._logger.info("<--")
        return return_list

    def __deepcopy__(self, memo):
        raise NotImplementedError()


class SelectionValidationExpConfsGenerator(ExpConfsGenerator):
    """
    Abstract base class for generator wrappers used to generate hp_selection or validation set.

    The classes which inherit from this class extract some input data from training set and put it in validation (if self._is_validation is True) or in hp_selection (if self._is_validation is False). How the split is performed depends on the single subclasses

    Attributes
    ----------
    _is_validation: bool

    Methods
    -------
    _get_selection_validation()
        Base static factory method for get_selection_generator and get_validation_generator

    get_selection_generator()
        Static factory method to instantiate subclasses

    get_validation_generator()
        Static factory method to instantiate subclasses
    """

    def __init__(self, campaign_configuration, seed, is_validation):
        """
        Since this is an abstract class, this constructor is only used by constructors of the subclasses

        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        is_validation: bool
            True if the subclass instance is created for validating, false if it is for hp selection
        """
        self._is_validation = is_validation
        super().__init__(campaign_configuration, seed)

    @staticmethod
    def _get_selection_validation_generator(campaign_configuration, seed, wrapped_generator, subclass_name, is_validation):
        """
        Static factory method to instantiate subclasses

        Parameters
        ----------
        campaign_configuration: dict of dict
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The generator to be wrapped

        subclass_name: str
            The name of the class to be generated

        is_validation: bool
            True if the instance to be created is for validating, false if it is for hp selection

        Returns
        -------
        SelectionValidationExpConfsGenerator
            The generated instance
        """
        logger = custom_logger.getLogger(__name__)
        if subclass_name == "All":
            return AllExpConfsGenerator(campaign_configuration, seed, wrapped_generator, is_validation)

        if subclass_name == "Extrapolation":
            if is_validation:
                # Split is performed as preprocessing step
                return ExtrapolationExpConfsGenerator(campaign_configuration, seed, wrapped_generator, is_validation)
            else:
                logger.error("Extrapolation cannot be used to perform hp_selection")
                sys.exit(1)
        elif subclass_name == "HoldOut":
            return HoldOutExpConfsGenerator(campaign_configuration, seed, wrapped_generator, is_validation)
        elif subclass_name == "KFold":
            return KFoldExpConfsGenerator(campaign_configuration, seed, campaign_configuration['General']['folds'],
                                          wrapped_generator, is_validation)
        else:
            logger.error("Unknown hp_selection/validation method %s", subclass_name)
            sys.exit(1)

    @staticmethod
    def get_selection_generator(campaign_configuration, seed, wrapped_generator, hp_selection_method):
        """
        Static factory method to instantiate subclasses which perform split of training into training and hp selection

        Parameters
        ----------
        campaign_configuration: dict of dict
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The generator to be wrapped

        hp_selection_method: str
            The name of the class to be generated

        Returns
        -------
        SelectionValidationExpConfsGenerator
            The generated instance
        """
        return SelectionValidationExpConfsGenerator._get_selection_validation_generator(campaign_configuration, seed, wrapped_generator, hp_selection_method, False)

    @staticmethod
    def get_validation_generator(campaign_configuration, seed, wrapped_generator, validation_method):
        """
        Static factory method to instantiate subclasses which perform split of training into training and validation

        Parameters
        ----------
        campaign_configuration: dict of dict
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The generator to be wrapped

        validation_method: str
            The name of the class to be generated

        Returns
        -------
        SelectionValidationExpConfsGenerator
            The generated instance
        """
        return SelectionValidationExpConfsGenerator._get_selection_validation_generator(campaign_configuration, seed, wrapped_generator, validation_method, True)

    def __deepcopy__(self, memo):
        raise NotImplementedError()


class XGBoostFeatureSelectionExpConfsGenerator(ExpConfsGenerator):
    """
    Wrapper class to perform feature selection based on XGBoost on training data

    Methods
    -------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, wrapped_generator):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The wrapped generator to be used with the selected features
        """
        assert campaign_configuration
        self._wrapped_generator = wrapped_generator
        super().__init__(campaign_configuration, seed)
        assert self._campaign_configuration

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        The generated experiment configurations are the ExperimentConfiguration generated by the wrapped generator, but with the features selected according to XGBoost

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by XGBoostFeatureSelectionExpConfsGenerator")
        xgboost_feature_selection = data_preparation.xgboost_feature_selection.XGBoostFeatureSelection(self._campaign_configuration, prefix)
        local_regression_inputs = copy.copy(regression_inputs)
        local_regression_inputs = xgboost_feature_selection.process(local_regression_inputs)
        ret_list = self._wrapped_generator.generate_experiment_configurations(prefix, local_regression_inputs)
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        assert self._campaign_configuration
        return XGBoostFeatureSelectionExpConfsGenerator(self._campaign_configuration, self._random_generator.random(), copy.deepcopy(self._wrapped_generator))


class AllExpConfsGenerator(SelectionValidationExpConfsGenerator):
    """
    Wraps a generator and pass to it the set of the regression inputs without modification

    Methods
    -------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, wrapped_generator, is_validation):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The generator to be wrapped

        is_validation: bool
            True if the instance to be created is for validating, false if it is for hp selection
        """
        self._wrapped_generator = wrapped_generator
        super().__init__(campaign_configuration, seed, is_validation)

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by AllExpConfsGenerator")
        local_prefix = copy.copy(prefix)
        local_prefix.append("All")
        local_regression_inputs = copy.copy(regression_inputs)

        if self._is_validation:
            local_regression_inputs.inputs_split["validation"] = local_regression_inputs.inputs_split["training"].copy()
        else:
            local_regression_inputs.inputs_split["hp_selection"] = local_regression_inputs.inputs_split["training"].copy()

        ret_list = self._wrapped_generator.generate_experiment_configurations(local_prefix, local_regression_inputs)
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        assert self._campaign_configuration
        return AllExpConfsGenerator(self._campaign_configuration, self._random_generator.random(), copy.deepcopy(self._wrapped_generator), self._is_validation)


class ExtrapolationExpConfsGenerator(SelectionValidationExpConfsGenerator):
    """
    Wraps a generator and pass to it regression inputs without modification.

    The actual extrapolation is indeed performed by the pre-processing step Extrapolation.

    Methods
    -------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, wrapped_generator, is_validation):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The generator to be wrapped

        is_validation: bool
            True if the instance to be created is for validating, false if it is for hp selection; in this class this parameter is not used, but it is not removed because of generality of the parent class
        """
        self._wrapped_generator = wrapped_generator
        super().__init__(campaign_configuration, seed, is_validation)

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by ExtrapolationExpConfsGenerator")
        local_prefix = copy.copy(prefix)
        local_prefix.append("Extrapolation")
        local_regression_inputs = copy.copy(regression_inputs)

        assert self._is_validation
        # Do nothing: validation set has already been built by the pre-processing step

        ret_list = self._wrapped_generator.generate_experiment_configurations(local_prefix, local_regression_inputs)
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        assert self._campaign_configuration
        return ExtrapolationExpConfsGenerator(self._campaign_configuration, self._random_generator.random(), copy.deepcopy(self._wrapped_generator), self._is_validation)


class HoldOutExpConfsGenerator(SelectionValidationExpConfsGenerator):
    """
    Wraps a generator and pass to it the regression_inputs removing data for hp_selection/validation

    How many data have to be removed from the training set is controlled by campaign_configuration['General']['hold_out_ratio']. The split is implemented in RandomSplitting pre-processing step which is executed before generating the ExperimentConfigurations.

    Methods
    -------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, wrapped_generator, is_validation):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The generator to be wrapped

        is_validation: bool
            True if the instance to be created is for validating, false if it is for hp selection
        """
        self._wrapped_generator = wrapped_generator
        super().__init__(campaign_configuration, seed, is_validation)

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by HoldOutExpConfsGenerator")
        local_prefix = copy.copy(prefix)
        local_prefix.append("HoldOut")
        local_regression_inputs = regression_inputs.copy()
        destination_set = "validation" if self._is_validation else "hp_selection"
        splitter = data_preparation.random_splitting.RandomSplitting(self._campaign_configuration,
                                                                     self._random_generator.random(), "training",
                                                                     destination_set)
        local_regression_inputs = splitter.process(local_regression_inputs)
        ret_list = self._wrapped_generator.generate_experiment_configurations(local_prefix, local_regression_inputs)
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        return HoldOutExpConfsGenerator(self._campaign_configuration, self._random_generator.random(),
                                        copy.deepcopy(self._wrapped_generator), self._is_validation)


class KFoldExpConfsGenerator(SelectionValidationExpConfsGenerator):
    """
    Wraps k instances of a generator with different training, hp_selection, and validation set

    Attributes
    ----------
    _k: integer
        The number of different folds to be used

    _kfold_generators: dict of integer: ExpConfsGenerator
        The wrapped generators which work on the single fold

    Methods
    ------
    generate_experiment_configurations()

    Generates the set of points to be evaluated
    """

    def __init__(self, campaign_configuration, seed, k, wrapped_generator, is_validation):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        k: integer
            The number of folds to be considered

        wrapped_generator: ExpConfsGenerator
            The wrapped generator to be duplicated and modified

        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used in random based activities
        """
        self._k = k

        self._kfold_generators = {}

        for fold in range(0, self._k):
            self._kfold_generators[fold] = copy.deepcopy(wrapped_generator)

        super().__init__(campaign_configuration, seed, is_validation)

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by KFoldExpConfsGenerator")
        assert prefix

        if len(regression_inputs.inputs_split["training"]) < self._k:
            self._logger.error("Too few samples to perform %d-Fold", self._k)
            sys.exit(1)

        return_list = []
        dataset_size = len(regression_inputs.inputs_split["training"])
        fold_size = int(dataset_size / self._k)
        all_training_idx = set(regression_inputs.inputs_split["training"])
        remaining = set(all_training_idx)

        for fold in range(0, self._k):
            self._logger.info("-->Generating experiments for fold %s", str(fold))
            assert prefix
            fold_prefix = copy.copy(prefix)
            assert fold_prefix
            fold_prefix.append("f" + str(fold))
            assert fold_prefix
            fold_regression_inputs = copy.copy(regression_inputs)

            if fold == self._k - 1:
                fold_testing_idx = remaining
            else:
                fold_testing_idx = set(self._random_generator.sample(list(remaining), fold_size))

            fold_training_idx = all_training_idx - fold_testing_idx
            remaining = remaining - fold_testing_idx
            fold_regression_inputs.inputs_split["training"] = list(fold_training_idx)
            second_set = "validation" if self._is_validation else "hp_selection"
            fold_regression_inputs.inputs_split[second_set] = list(fold_testing_idx)
            return_list.extend(self._kfold_generators[fold].generate_experiment_configurations(fold_prefix, fold_regression_inputs))
            self._logger.info("<--")

        self._logger.info("<--")
        return return_list

    def __deepcopy__(self, memo):
        return KFoldExpConfsGenerator(self._campaign_configuration, self._random_generator.random(), self._k,
                                      self._kfold_generators[0], self._is_validation)


class NormalizationExpConfsGenerator(ExpConfsGenerator):
    """
    Wraps a generator and pass to it regression inputs normalized according to the value of the training set

    Methods
    -------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, wrapped_generator):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used to initialize the random generator

        wrapped_generator: ExpConfsGenerator
            The wrapped generator which generate the ExperimentConfigurations
        """
        self._wrapped_generator = wrapped_generator
        super().__init__(campaign_configuration, seed)

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by NormalizationExpConfsGenerator")
        normalizer = data_preparation.normalization.Normalization(self._campaign_configuration)
        local_regression_inputs = copy.copy(regression_inputs)
        local_regression_inputs = normalizer.process(local_regression_inputs)
        ret_list = self._wrapped_generator.generate_experiment_configurations(prefix, local_regression_inputs)
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        return NormalizationExpConfsGenerator(self._campaign_configuration, self._random_generator.random(),
                                              copy.deepcopy(self._wrapped_generator))


class RandomExpConfsGenerator(ExpConfsGenerator):
    """
    Wraps an experiment configuration generator randomly picking n experiment configurations

    Attributes
    ----------
    _experiment_configurations_number: integer
        The number of experiment configurations to be returned

    _wrapped_generator: ExpConfsGenerator
        The wrapped generator to be used

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of experiment configurations to be evaluated
    """

    def __init__(self, campaign_configuration, seed, experiment_configurations_number, wrapped_generator):
        """
        Parameters
        ----------
        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used in random based activities

        experiment_configurations_number: integer
            The number of experiment configurations to select

        wrapped_generator: ExpConfsGenerator
            The wrapped generator
        """
        super().__init__(campaign_configuration, seed)

        self._experiment_configurations_number = experiment_configurations_number
        self._wrapped_generator = wrapped_generator

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generate the experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used in the computation of the signatures of the experiment configurations

        regression_inputs: RegressionInputs
            The input of the regression problem to be used in experiment configurations generated by self

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        # TODO  call wrapped generator and randomly pick n experiment configurations

    def __deepcopy__(self, memo):
        raise NotImplementedError()




class SFSExpConfsGenerator(ExpConfsGenerator):
    """
    Generator for using SFS selection

    The class wraps an ExpConfsGenerator to apply SFS for performing feature selection on the input dataset; the SFS is used to filter the columns of the dataset which is passed as argument to the wrapped generator

    Attributes
    ----------
    wrapped_generator: ExpConfsGenerator
        The ExpConfsGenerator used with the feature selector

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of expriment configurations to be evaluated
    """

    def __init__(self, wrapped_generator, campaign_configuration, seed):
        """
        Parameters
        ----------
        wrapped_generator: ExpConfsGenerator
            The ExpConfsGenerator to be used coupled with the current feature selector

        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used in random based activities
        """
        super().__init__(campaign_configuration, seed)
        self._wrapped_generator = wrapped_generator

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generates the set of experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used for the signature of the ExperimentConfiguration generated by this generator

        regression_inputs: RegressionInputs
            The data to be used by the ExperimentConfiguration

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by SFSExpConfsGenerator")
        internal_list = self._wrapped_generator.generate_experiment_configurations(prefix, regression_inputs)
        ret_list = []
        for wrapped_point in internal_list:
            ret_list.append(wec.SFSExperimentConfiguration(self._campaign_configuration, copy.deepcopy(regression_inputs), prefix, wrapped_point))
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        return SFSExpConfsGenerator(copy.deepcopy(self._wrapped_generator), self._campaign_configuration, self._random_generator.random())


class HyperoptExpConfsGenerator(ExpConfsGenerator):
    """
    Generator for using hyperparameter selection via the Hyperopt library

    The class wraps an ExpConfsGenerator to use Hyperopt for hyperparameter selection of its own wrapped generator

    Attributes
    ----------
    wrapped_generator: ExpConfsGenerator
        The ExpConfsGenerator that requires tuning via Hyperopt

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of expriment configurations to be evaluated
    """
    def __init__(self, wrapped_generator, campaign_configuration, seed):
        """
        Parameters
        ----------
        wrapped_generator: ExpConfsGenerator
            The ExpConfsGenerator to be used coupled with the current feature selector

        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used in random based activities
        """
        super().__init__(campaign_configuration, seed)
        self._wrapped_generator = wrapped_generator

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generates the set of experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used for the signature of the ExperimentConfiguration generated by this generator

        regression_inputs: RegressionInputs
            The data to be used by the ExperimentConfiguration

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by HyperoptExpConfsGenerator")
        internal_list = self._wrapped_generator.generate_experiment_configurations(prefix, regression_inputs)
        ret_list = []
        for wrapped_point in internal_list:
            ret_list.append(wec.HyperoptExperimentConfiguration(self._campaign_configuration, copy.deepcopy(regression_inputs), prefix, wrapped_point))
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        return HyperoptExpConfsGenerator(copy.deepcopy(self._wrapped_generator), self._campaign_configuration, self._random_generator.random())


class HyperoptSFSExpConfsGenerator(ExpConfsGenerator):
    """
    Generator for using both Hyperopt tuning and SFS selection

    Attributes
    ----------
    wrapped_generator: ExpConfsGenerator
        The ExpConfsGenerator used with the feature selector

    Methods
    ------
    generate_experiment_configurations()
        Generates the set of expriment configurations to be evaluated
    """
    def __init__(self, wrapped_generator, campaign_configuration, seed):
        """
        Parameters
        ----------
        wrapped_generator: ExpConfsGenerator
            The ExpConfsGenerator to be used coupled with the current feature selector

        campaign_configuration: dict of str: dict of str: str
            The set of options specified by the user though command line and campaign configuration files

        seed: integer
            The seed to be used in random based activities
        """
        super().__init__(campaign_configuration, seed)
        self._wrapped_generator = wrapped_generator

    def generate_experiment_configurations(self, prefix, regression_inputs):
        """
        Generates the set of experiment configurations to be evaluated

        Parameters
        ----------
        prefix: list of str
            The prefix to be used for the signature of the ExperimentConfiguration generated by this generator

        regression_inputs: RegressionInputs
            The data to be used by the ExperimentConfiguration

        Returns
        -------
        list of ExperimentConfiguration
            a list of the experiment configurations to be evaluated
        """
        self._logger.info("-->Generating experiments by HyperoptSFSExpConfsGenerator")
        internal_list = self._wrapped_generator.generate_experiment_configurations(prefix, regression_inputs)
        ret_list = []
        for wrapped_point in internal_list:
            ret_list.append(wec.HyperoptSFSExperimentConfiguration(self._campaign_configuration, copy.deepcopy(regression_inputs), prefix, wrapped_point))
        self._logger.info("<--")
        return ret_list

    def __deepcopy__(self, memo):
        return HyperoptSFSExpConfsGenerator(copy.deepcopy(self._wrapped_generator), self._campaign_configuration, self._random_generator.random())
